#ifndef __VISTPC_H__
#define __VISTPC_H__

#include "TBBQManager.h"

// C++

#include <map>
#include <string>
#include <vector>

// ROOT

class TGeoMaterial;
class TGeoMedium;
class TGeoVolume;

// Eve

class TEveBoxSet;
class TEveElement;
class TEvePointSet;
class TEveQuadSet;
class TEveStraightLineSet;

// GEAR

namespace gear {
   class GearMgr;
   class TPCModule;
   class TPCParameters;
}

#ifndef __CINT__
#include "gearimpl/Vectors.h"
#else
namespace gear {
   class Vector2D;
}
#endif

// LCIO

#include "EVENT/TrackerHit.h"

namespace EVENT {
   class LCEvent;
   class TrackerHit;
}

namespace IMPL {
   class LCCollectionVec;
}

/*!
 *  \brief
 *  Main visualization class. It is used to draw GEAR detector geometries and
 *  LCIO data. It is a self-contained class that depends only on ROOT, TEve,
 *  GEAR and LCIO. This means you can use this class outside and independent
 *  of BBQ.
 *
 *  @author Ching Bon Lam
 */

class VisTPC {
public:
   
   //---------------------------------------------------------------------------
   // Data types and structures

   enum LCCollectionType_enum {
      kUNDEFINEDTYPE,
      kPULSE,
      kHIT,
      kTRACK
   };
   
   /*!
    *  LeftRight_enum is only used for the method
    *  testLeftRightNeighbourWithPads(). It specifies which GEAR function to
    *  call: getLeftNeighbour() or getRightNeighbour().
    */
   
   enum LeftRight_enum {
      kLEFT,
      kRIGHT
   };

   /*!
       \brief
       The PadMetric is used for holding the geometric properties of a pad.
       
       The variable #coordinate_system holds the value
       gear::PadRowLayout2D::CARTESIAN or gear::PadRowLayout2D::POLAR.
    
       \verbatim
    
       Corner layout
    
        y | (6,7)________(4,5)
          |     |        |
          |     |        |
          |     |________|
          | (0,1)        (2,3)
          |______________________
                               x
       \endverbatim
       
       @author Ching Bon Lam
   */

   struct PadMetric {
      int coordinate_system;
      
      float height;
      float width;
      
      float center_x;
      float center_y;
      
      float center_r;
      float center_phi;
      
      float corners_polar[8];
      float corners_cartesian[8];
   };
   
   //---------------------------------------------------------------------------
   // Constructor

   VisTPC(gear::GearMgr * aGearMgr = 0);
   ~VisTPC() {};
   
   //---------------------------------------------------------------------------
   /// @name Setters/Getters
   
   //@{
   
   float getReadoutFrequency()               {return readoutFrequency;}
   float getDriftVelocity()                  {return driftVelocity;}
   float getMaxDriftLength()                 {return maxDriftLength;}
   int   getMaxPoints()                      {return maxPoints;}
   int   getOption(std::string& option)      {
      std::map<std::string, int>::iterator it = options.find(option);
      
      if (it == options.end())
         throw "Option not found";
      
      return options[option];
   }
   
   void setReadoutFrequency(float frequency) {readoutFrequency = frequency;}
   void setDriftVelocity(float velocity)     {driftVelocity = velocity;}
   void setMaxDriftLength(float length)      {maxDriftLength = length;}
   void setMaxPoints(int points)             {maxPoints = points;}
   void setOption(std::string option, int value) {options[option] = value;}
   
   //@}

   //---------------------------------------------------------------------------
   /*! @name GEAR methods
    *
    *  These methods are for drawing the detector and modules as described in
    *  the GEAR file.
    *  
    *  When the "parent" is NULL, the visualized data will also be added to the
    *  global scene. When the "parent" is not NULL, the visualized data will be
    *  added to "parent".
    */

   //@{
   
   void drawTPC();
   
   void drawModule(gear::TPCModule* aModule, TEveElement* parent = NULL);
   void drawModules(TEveElement* parent = NULL);
   
   void drawPads(gear::TPCModule* aModule, TEveElement* parent = NULL);
   
   //@}
   
   //---------------------------------------------------------------------------
   /*! @name LCIO methods
    *
    *  These methods are for visualizing the data in the LCIO file.
    *
    *  When there is no "parent" in the argument list, the visualized data will
    *  be added to the event scene. When the "parent" is NULL, the visualized
    *  data will also be added to the event scene. When the "parent" is not
    *  NULL, the visualized data will be added to "parent".
    *
    *  When a TEveElement* value is returned, it can be used to set properties
    *  in the caller function.
    *
    *  This inconsistency is due to historical/evolutionary reasons. Both styles
    *  has pro's and con's.
    */
   
   //@{
   
   void         drawEvent(EVENT::LCEvent* event, std::vector<std::string> &selectedEntries, TEveElement* parent = NULL, const char* name = "Event");
   void         drawEvent(EVENT::LCEvent* event, TList &selectedEntries,                    TEveElement* parent = NULL, const char* name = "Event");
   void         drawHit(EVENT::TrackerHit* hit, TEvePointSet* parent, int wantTooltip = false);
   TEveElement* drawHits(IMPL::LCCollectionVec* hits, int wantTooltip = true);
   TEveElement* drawHits(const EVENT::TrackerHitVec& hits, int wantTooltip = true);
   TEveElement* drawPulses(IMPL::LCCollectionVec* pulses);
   TEveElement* drawTracks(IMPL::LCCollectionVec* tracks, TEveElement* parent = NULL);
   
   //@}
   
   //---------------------------------------------------------------------------
   /*! @name Helper methods
    *
    *  These methods are more generic drawing functions that are used by all
    *  other methods.
    */

   //@{

   void drawCircularPolygon(TEveStraightLineSet* ls, float r, int nDivisions, float z);
   void drawCircularPolygon(TEveStraightLineSet* ls, float r, float phi_min, float phi_max, int nDivisions, float z);

   void drawPadBox(const PadMetric& pad, float z_front, float z_back, TEveBoxSet* bs);
   void drawPadQuad(const PadMetric& pad, float z, TEveQuadSet* qs);
   
   const LCCollectionType_enum getCollectionType(const std::string&);

   /*! \brief
    *  Get the pad geometric properties in the global coordinate system of the
    *  detector.
    */

   void getGlobalPadMetric(const gear::TPCModule * aModule, const int padIndex, PadMetric& pad);
   
   /*! \brief
    *  Get the pad geometric properties in the local coordinate system of the
    *  module.
    */
   
   void getLocalPadMetric(const gear::TPCModule * aModule, const int padIndex, PadMetric& pad);
   
   /*! \brief
    *  Transform a TEveElement to the global coordinate system of the detector
    *  using OpenGL matrix transformation. The transformation matrix is
    *  constructed using the information of the TPCModule.
    */
   
   void toGlobal(const gear::TPCModule * aModule, TEveElement* element);
   
   //@}

   //---------------------------------------------------------------------------
   /*! @name Tests of GEAR functionality
    *
    *  These methods are for visualization of some important GEAR functions.
    */
   
   
   //@{
   
   void drawTwoConnectingPads(const gear::TPCModule& aModule, TEveElement* parent, int pad1, int pad2, LeftRight_enum leftright);
   void drawNearestPad(const gear::TPCModule& aModule, TEveElement* parent, int padIndex, const gear::Vector2D& position, int i);

   void testLeftRightNeighbour(int moduleID = 0);
   void testLeftRightNeighbourWithPads(int moduleID, const std::vector<int>& pads, LeftRight_enum leftright);
   
   void testNearestPad(int moduleID = 0);
   void testNearestPadWithPoints(int moduleID, const std::vector<gear::Vector2D>& points);
   
   void testNearestPadGlobal();
   void testNearestPadGlobalWithPoints(const std::vector<gear::Vector2D>& points);
   
   //@}

   /// A function to update the vistpc with information from a new gear file
   void updateGearMgr( gear::GearMgr * aGearMgr );

   //---------------------------------------------------------------------------
   // Variables
protected:

   int debug;

   /// @name GEAR properties
   //@{

   gear::GearMgr *             gearMgr;
   gear::TPCParameters const * tpcParameters;

   //@}

   /*! @name TGeo properties
    *  This is used to draw the TPC ( drawTPC() ) and isn't used anywhere else.
    */
   
   //@{

   std::map<std::string, TGeoMaterial*> materials;
   std::map<std::string, TGeoMedium*>   media;
   std::map<std::string, TGeoVolume*>   volumes;
   
   //@}
   
   /*! @name "Hard coded" detector properties
    *  These properties should be taken from GEAR or LCIO, but aren't for some
    *  reason.
    */
   
   //@{
   
   float readoutFrequency;
   float driftVelocity;
   
   /*! \brief
    *  This is used for the z-position of the endcap/modules.
    *  
    *  Yes, this is not good. It is called maxDriftLength because it gets its
    *  value from tpcParameters->getMaxDriftLength().
    */
   
   float maxDriftLength;
   
   //@}

   /*! @name Visualisation options
    *  These options allow to turn on or off certain features.
    */
   
   //@{
   
   std::map<std::string, int> options;
   
   //@}
   
   int maxPoints;
   
   float outerFieldCageRadius; ///< Radius of the outer field cage cylinder
   float innerFieldCageRadius; ///< Radius of the inner field cage cylinder

private:
   ClassDef(VisTPC, 1);
};

#endif
